@@ -217,6 +217,9 @@ static NTSTATUS smbd_smb2_tree_connect(struct smbd_smb2_request *req,
 	bool encryption_required = req->session->global->encryption_flags & SMBXSRV_ENCRYPTION_REQUIRED;
 	bool guest_session = false;
 	bool require_signed_tcon = false;
+	uint32_t session_global_id;
+	char *share_name = NULL;
+	uint8_t encryption_flags = 0;
 
 	*disconnect = false;
 
@@ -328,17 +331,39 @@ static NTSTATUS smbd_smb2_tree_connect(struct smbd_smb2_request *req,
 		}
 	}
 
-	/* create a new tcon as child of the session */
-	status = smb2srv_tcon_create(req->session, now, &tcon);
-	if (!NT_STATUS_IS_OK(status)) {
-		return status;
-	}
-
 	if (encryption_desired) {
-		tcon->global->encryption_flags |= SMBXSRV_ENCRYPTION_DESIRED;
+		encryption_flags |= SMBXSRV_ENCRYPTION_DESIRED;
 	}
 	if (encryption_required) {
-		tcon->global->encryption_flags |= SMBXSRV_ENCRYPTION_REQUIRED;
+		encryption_flags |= SMBXSRV_ENCRYPTION_REQUIRED;
+	}
+
+	session_global_id = req->session->global->session_global_id;
+	share_name = lp_servicename(talloc_tos(), lp_sub, snum);
+	if (share_name == NULL) {
+		return NT_STATUS_NO_MEMORY;
+	}
+
+	if ((lp_max_connections(snum) > 0)
+	    && (count_current_connections(lp_const_servicename(snum), true) >=
+		lp_max_connections(snum))) {
+
+		DBG_WARNING("Max connections (%d) exceeded for [%s][%s]\n",
+			  lp_max_connections(snum),
+			  lp_const_servicename(snum), share_name);
+		TALLOC_FREE(share_name);
+		return NT_STATUS_INSUFFICIENT_RESOURCES;
+	}
+
+	/* create a new tcon as child of the session */
+	status = smb2srv_tcon_create(req->session,
+				     session_global_id,
+				     encryption_flags,
+				     share_name,
+				     now, &tcon);
+	TALLOC_FREE(share_name);
+	if (!NT_STATUS_IS_OK(status)) {
+		return status;
 	}
 
 	compat_conn = make_connection_smb2(req,
@@ -350,27 +375,10 @@ static NTSTATUS smbd_smb2_tree_connect(struct smbd_smb2_request *req,
 		return status;
 	}
 
-	tcon->global->share_name = lp_servicename(tcon->global,
-						  lp_sub,
-						  SNUM(compat_conn));
-	if (tcon->global->share_name == NULL) {
-		conn_free(compat_conn);
-		TALLOC_FREE(tcon);
-		return NT_STATUS_NO_MEMORY;
-	}
-	tcon->global->session_global_id =
-		req->session->global->session_global_id;
-
 	tcon->compat = talloc_move(tcon, &compat_conn);
 
 	tcon->status = NT_STATUS_OK;
 
-	status = smbXsrv_tcon_update(tcon);
-	if (!NT_STATUS_IS_OK(status)) {
-		TALLOC_FREE(tcon);
-		return status;
-	}
-
 	if (IS_PRINT(tcon->compat)) {
 		*out_share_type = SMB2_SHARE_TYPE_PRINT;
 	} else if (IS_IPC(tcon->compat)) {
@@ -413,6 +421,84 @@ static NTSTATUS smbd_smb2_tree_connect(struct smbd_smb2_request *req,
 		*out_share_flags |= SMB2_SHAREFLAG_ENCRYPT_DATA;
 	}
 
+	/*
+	 * For disk shares we can change the client
+	 * behavior on a cluster...
+	 */
+	if (conn->protocol >= PROTOCOL_SMB3_00 &&
+	    *out_share_type == SMB2_SHARE_TYPE_DISK)
+	{
+		bool persistent = false; /* persistent handles not implemented yet */
+		bool cluster = lp_clustering();
+		bool scaleout = cluster;
+		bool witness = cluster && !lp_rpc_start_on_demand_helpers();
+		bool asymmetric = false; /* shares are symmetric by default */
+		bool announce;
+
+		/*
+		 * In a ctdb cluster shares are continuously available,
+		 * but windows clients mix this with the global persistent
+		 * handles support.
+		 *
+		 * Persistent handles are requested if
+		 * SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY is present
+		 * even without SMB2_CAP_PERSISTENT_HANDLES.
+		 *
+		 * And SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY is
+		 * required for SMB2_SHARE_CAP_CLUSTER to have
+		 * an effect.
+		 *
+		 * So we better don't announce this by default
+		 * until we support persistent handles.
+		 */
+		announce = lp_parm_bool(SNUM(tcon->compat),
+					"smb3 share cap",
+					"CONTINUOUS AVAILABILITY",
+					persistent);
+		if (announce) {
+			*out_capabilities |= SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY;
+		}
+
+		/*
+		 * ctdb clusters are always scale out...
+		 */
+		announce = lp_parm_bool(SNUM(tcon->compat),
+					"smb3 share cap",
+					"SCALE OUT",
+					scaleout);
+		if (announce) {
+			*out_capabilities |= SMB2_SHARE_CAP_SCALEOUT;
+		}
+
+		/*
+		 * We support the witness service when ctdb is active
+		 */
+		announce = lp_parm_bool(SNUM(tcon->compat),
+					"smb3 share cap",
+					"CLUSTER",
+					witness);
+		if (announce) {
+			*out_capabilities |= SMB2_SHARE_CAP_CLUSTER;
+		}
+
+		/*
+		 * Shares in a ctdb cluster are symmetric by design.
+		 *
+		 * But it might be useful to let the client use
+		 * an isolated transport and witness registration for the
+		 * specific share.
+		 */
+		if (conn->protocol >= PROTOCOL_SMB3_02) {
+			announce = lp_parm_bool(SNUM(tcon->compat),
+						"smb3 share cap",
+						"ASYMMETRIC",
+						asymmetric);
+		}
+		if (announce) {
+			*out_capabilities |= SMB2_SHARE_CAP_ASYMMETRIC;
+		}
+	}
+
 	*out_maximal_access = tcon->compat->share_access;
 
 	*out_tree_id = tcon->global->tcon_wire_id;
